# Copyright (c) 2025, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""
Readers for ASCAT Level 1b in HDF5 format.
"""

from collections import defaultdict
from collections import OrderedDict

import h5py
import numpy as np
import xarray as xr

from ascat.utils import mask_dtype_nans
from ascat.read_native import AscatFile
from ascat.read_native.eps_native import set_flags

class AscatL1bHdf5File(AscatFile):
    """
    Class reading ASCAT Level 1b file in HDF5 format.
    """

    def _read(self, filename, generic=False, to_xarray=False):
        """
        Read one ASCAT Level 1b HDF5 file.

        Parameters
        ----------
        generic : bool, optional
            "True" reading and converting into generic format or
            "False" reading original field names (default: False).
        to_xarray : bool, optional
            "True" return data as xarray.Dataset
            "False" return data as numpy.ndarray (default: False).

        Returns
        -------
        ds : dict
            ASCAT data.
        metadata : dict
            Metadata.
        """
        data = {}
        metadata = {}

        root = "U-MARF/EPS/ASCA_SZF_1B/"
        mdr_path = root + "DATA/MDR_1B_FULL_ASCA_Level_1_ARRAY_000001"
        mdr_descr_path = root + "DATA/MDR_1B_FULL_ASCA_Level_1_DESCR"
        metadata_path = root + "METADATA"

        with h5py.File(filename, mode="r") as fid:
            mdr = fid[mdr_path]
            mdr_descr = fid[mdr_descr_path]
            mdr_metadata = fid[metadata_path]
            var_names = list(mdr.dtype.names)

            for var_name in var_names:
                data[var_name.lower()] = mdr[var_name]

                if var_name.encode() in mdr_descr["EntryName"]:
                    pos = mdr_descr["EntryName"] == var_name.encode()
                    scale = mdr_descr["Scale Factor"][pos][0].decode()

                    if scale != "n/a":
                        data[var_name.lower()] = (data[
                            var_name.lower()] / (10. ** float(scale))).astype(
                                np.float32)

            fields = ["SPACECRAFT_ID", "ORBIT_START",
                      "PROCESSOR_MAJOR_VERSION", "PROCESSOR_MINOR_VERSION",
                      "FORMAT_MAJOR_VERSION", "FORMAT_MINOR_VERSION"]

            for f in fields:
                pos = np.char.startswith(
                    mdr_metadata["MPHR/MPHR_TABLE"]["EntryName"], f.encode())
                var = mdr_metadata["MPHR/MPHR_TABLE"]["EntryValue"][
                    pos][0].decode()

                if f == "SPACECRAFT_ID":
                    var = var[-1]

                metadata[f.lower()] = int(var)

        # modify longitudes [0, 360] to [-180, 180]
        mask = data["longitude_full"] > 180
        data["longitude_full"][mask] += -360.

        data["time"] = np.datetime64("2000-01-01") + data[
            "utc_localisation-days"].astype("timedelta64[D]") + data[
                "utc_localisation-milliseconds"].astype("timedelta64[ms]")

        # modify azimuth angles to [0, 360]
        if "azi_angle_full" in var_names:
            mask = data["azi_angle_full"] < 0
            data["azi_angle_full"][mask] += 360

        rename_coords = {"longitude_full": ("lon", np.float32),
                         "latitude_full": ("lat", np.float32)}

        for var_name, (new_name, new_dtype) in rename_coords.items():
            data[new_name] = data.pop(var_name).astype(new_dtype)

        if generic:
            data = conv_hdf5l1b_generic(data, metadata)

        # 1 Left Fore Antenna, 2 Left Mid Antenna 3 Left Aft Antenna
        # 4 Right Fore Antenna, 5 Right Mid Antenna, 6 Right Aft Antenna
        left_beams = ["lf-vv", "lm-vv", "la-vv"]
        right_beams = ["rf-vv", "rm-vv", "ra-vv"]

        all_beams = left_beams + right_beams

        ds = OrderedDict()

        for i, beam in enumerate(all_beams):

            subset = data["beam_number"] == i+1
            metadata["beam_number"] = i+1
            metadata["beam_name"] = beam

            # convert dict to xarray.Dataset or numpy.ndarray
            if to_xarray:
                sub_data = {}
                for var_name in data.keys():

                    if var_name == "beam_number" and generic:
                        continue

                    if len(data[var_name].shape) == 1:
                        dim = ["obs"]
                    elif len(data[var_name].shape) == 2:
                        dim = ["obs", "echo"]

                    sub_data[var_name] = (dim, data[var_name][subset])

                coords = {}
                coords_fields = ["lon", "lat", "time"]

                for cf in coords_fields:
                    coords[cf] = sub_data.pop(cf)

                ds[beam] = xr.Dataset(sub_data, coords=coords,
                                      attrs=metadata)
                if generic:
                    data = mask_dtype_nans(data)
            else:
                # collect dtype info
                dtype = []
                for var_name in data.keys():

                    if len(data[var_name][subset].shape) == 1:
                        dtype.append(
                            (var_name, data[var_name][subset].dtype.str))
                    elif len(data[var_name][subset].shape) > 1:
                        dtype.append((var_name, data[var_name][
                            subset].dtype.str, data[var_name][
                                subset].shape[1:]))

                ds[beam] = np.empty(
                    data["time"][subset].size, dtype=np.dtype(dtype))

                for var_name in data.keys():
                    if var_name == "beam_number" and generic:
                        continue
                    ds[beam][var_name] = data[var_name][subset]

        return ds, metadata

    def _merge(self, data):
        """
        Merge data.

        Parameters
        ----------
        data : list
            List of array.

        Returns
        -------
        data : numpy.ndarray
            Data.
        """
        left_beams = ["lf-vv", "lm-vv", "la-vv"]
        right_beams = ["rf-vv", "rm-vv", "ra-vv"]
        all_beams = left_beams + right_beams

        metadata = {}

        if isinstance(data[0], tuple):
            data, metadata = zip(*data)
        merged_data = defaultdict(list)
        for beam in all_beams:
            for d in data:
                merged_data[beam].append(d.pop(beam))
            if isinstance(merged_data[beam][0], xr.Dataset):
                merged_data[beam] = xr.concat(merged_data[beam],
                                              dim="obs",
                                              combine_attrs="drop_conflicts")
            else:
                merged_data[beam] = np.hstack(merged_data[beam])

        merged_data = (merged_data, metadata)

        return merged_data

class AscatL1bHdf5FileGeneric(AscatL1bHdf5File):
    """
    The same as AscatL1bHdf5File but with generic=True by default.
    """
    def _read(self, filename, generic=True, to_xarray=False, **kwargs):
        return super()._read(filename, generic=generic, to_xarray=to_xarray, **kwargs)


def conv_hdf5l1b_generic(data, metadata):
    """
    Rename and convert data types of dataset.

    Parameters
    ----------
    data : dict of numpy.ndarray
        Original dataset.
    metadata : dict
        Metadata.

    Returns
    -------
    data : dict of numpy.ndarray
        Converted dataset.
    """
    # convert spacecraft_id to internal sat_id
    sat_id = np.array([4, 3, 5])
    metadata["sat_id"] = sat_id[metadata["spacecraft_id"]-1]

    # compute ascending/descending direction
    data["as_des_pass"] = (
        data["sat_track_azi"] < 270).astype(np.uint8)

    flags = {"flagfield_rf1": np.tile(data["flagfield_rf1"], 192),
             "flagfield_rf2": np.tile(data["flagfield_rf2"], 192),
             "flagfield_pl": np.tile(data["flagfield_pl"], 192),
             "flagfield_gen1": data["flagfield_gen1"].flatten(),
             "flagfield_gen2": data["flagfield_gen2"].flatten()}

    data["f_usable"] = set_flags(flags)
    data["f_usable"] = data["f_usable"].reshape(-1, 192)

    data["swath_indicator"] = np.int8(data["beam_number"].flatten() > 3)

    skip_fields = ["utc_localisation-days", "utc_localisation-milliseconds",
                   "degraded_inst_mdr", "degraded_proc_mdr", "flagfield_rf1",
                   "flagfield_rf2", "flagfield_pl", "flagfield_gen1",
                   "flagfield_gen2"]

    gen_fields_lut = {"inc_angle_full": ("inc", np.float32),
                      "azi_angle_full": ("azi", np.float32),
                      "sigma0_full": ("sig", np.float32)}

    for var_name in skip_fields:
        if var_name in data:
            data.pop(var_name)

    num_cells = data["lat"].shape[1]

    for var_name in data.keys():
        if len(data[var_name].shape) == 1:
            data[var_name] = np.repeat(data[var_name], num_cells)
        if len(data[var_name].shape) == 2:
            data[var_name] = data[var_name].flatten()

        if var_name in gen_fields_lut.items():
            new_name = gen_fields_lut[var_name][0]
            new_dtype = gen_fields_lut[var_name][1]
            data[new_name] = data.pop(var_name).astype(new_dtype)

    return data



# Copyright (c) 2025, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Readers for ASCAT Level 1b and Level 2 data in NetCDF format.
"""

import os
from datetime import datetime
from datetime import timedelta

import netCDF4
import numpy as np
import xarray as xr

from ascat.utils import tmp_unzip
from ascat.utils import daterange
from ascat.utils import mask_dtype_nans
from ascat.utils import uint8_nan
from ascat.utils import float32_nan
from ascat.file_handling import ChronFiles
from ascat.read_native import AscatFile


def read_nc(filename, generic, to_xarray, skip_fields, gen_fields_lut):
    """
    Read NetCDF file.

    Parameters
    ----------
    filename : str
        Filename.
    generic : bool
        'True' reading and converting into generic format or
        'False' reading original field names.
    to_xarray : bool
        'True' return data as xarray.Dataset
        'False' return data as numpy.ndarray.
    skip_fields : list
        Variables to skip.
    gen_fields_lut : dict
        Conversion look-up table for generic names.

    Returns
    -------
    data : xarray.Dataset or numpy.ndarray
        ASCAT data.
    metadata : dict
        Metadata.
    """
    data = {}
    metadata = {}

    with netCDF4.Dataset(filename) as fid:

        if hasattr(fid, 'platform'):
            metadata['platform_id'] = fid.platform[2:]
        elif hasattr(fid, 'platform_long_name'):
            metadata['platform_id'] = fid.platform_long_name[2:]

        metadata['orbit_start'] = fid.start_orbit_number
        metadata['processor_major_version'] = fid.processor_major_version
        metadata['product_minor_version'] = fid.product_minor_version
        metadata['format_major_version'] = fid.format_major_version
        metadata['format_minor_version'] = fid.format_minor_version
        metadata['filename'] = os.path.basename(filename)

        num_rows = fid.dimensions['numRows'].size
        num_cells = fid.dimensions['numCells'].size

        dtype = []
        for var_name in fid.variables.keys():

            if var_name in ['sigma0']:
                continue

            if generic and var_name in skip_fields:
                continue

            if generic and var_name in gen_fields_lut:
                new_var_name = gen_fields_lut[var_name][0]
                fill_value = gen_fields_lut[var_name][2]
            else:
                new_var_name = var_name
                fill_value = None

            var_data = fid.variables[var_name][:].filled(fill_value)

            if var_name == 'azi_angle_trip':
                var_data[(var_data < 0) & (var_data != fill_value)] += 360

            if len(fid.variables[var_name].shape) == 1:
                var_data = var_data.repeat(num_cells)
            elif len(fid.variables[var_name].shape) == 2:
                var_data = var_data.flatten()
            elif len(fid.variables[var_name].shape) == 3:
                var_data = var_data.reshape(-1, 3)
            else:
                raise RuntimeError('Unknown dimension')

            if var_name == 'utc_line_nodes':
                var_data = var_data.astype('timedelta64[s]') + np.datetime64(
                    '2000-01-01')

            data[new_var_name] = var_data

            if len(var_data.shape) == 1:
                dtype.append((new_var_name, var_data.dtype.str))
            elif len(var_data.shape) > 1:
                dtype.append(
                    (new_var_name, var_data.dtype.str, var_data.shape[1:]))

    num_records = num_rows * num_cells
    coords_fields = ['lon', 'lat', 'time']

    if generic:
        sat_id = np.array([0, 4, 3, 5], dtype=np.uint8)
        data['sat_id'] = np.zeros(num_records, dtype=np.uint8) + sat_id[int(
            metadata['platform_id'])]
        dtype.append(('sat_id', np.uint8))

        n_records = data['lat'].shape[0]
        n_lines = n_records // num_cells

        data['node_num'] = np.tile((np.arange(num_cells) + 1), n_lines)
        dtype.append(('node_num', np.uint8))

        data['line_num'] = np.arange(n_lines).repeat(num_cells)
        dtype.append(('line_num', np.int32))

    if to_xarray:
        for k in data.keys():
            if len(data[k].shape) == 1:
                dim = ['obs']
            elif len(data[k].shape) == 2:
                dim = ['obs', 'beam']

            data[k] = (dim, data[k])

        coords = {}
        for cf in coords_fields:
            coords[cf] = data.pop(cf)

        data = xr.Dataset(data, coords=coords, attrs=metadata)
        if generic:
            data = mask_dtype_nans(data)
    else:
        ds = np.empty(num_records, dtype=np.dtype(dtype))
        for k, v in data.items():
            ds[k] = v
        data = ds

    return data, metadata

class AscatL1bNcFile(AscatFile):
    """
    Read ASCAT Level 1b file in NetCDF format.
    """

    def __init__(self, filename, **kwargs):
        """
        Initialize AscatL1bNcFile.

        Parameters
        ----------
        filename : str
            Filename.
        """
        super().__init__(filename, **kwargs)
        for i, fname in enumerate(self.filenames):
            if os.path.splitext(fname)[1] == '.gz':
                self.filenames[i] = tmp_unzip(fname)

    def _read(self, filename, generic=False, to_xarray=False):
        """
        Read one ASCAT Level 1b NetCDF4 file.

        Parameters
        ----------
        generic : bool, optional
            'True' reading and converting into generic format or
            'False' reading original field names (default: False).
        to_xarray : bool, optional
            'True' return data as xarray.Dataset
            'False' return data as numpy.ndarray (default: False).

        Returns
        -------
        data : xarray.Dataset or numpy.ndarray
            ASCAT data.
        metadata : dict
            Metadata.
        """
        gen_fields_lut = {
            'longitude': ('lon', np.float32, None),
            'latitude': ('lat', np.float32, None),
            'utc_line_nodes': ('time', np.float32, None),
            'inc_angle_trip': ('inc', np.float32, float32_nan),
            'azi_angle_trip': ('azi', np.float32, float32_nan),
            'sigma0_trip': ('sig', np.float32, float32_nan),
            'kp': ('kp', np.float32, float32_nan),
            'f_kp': ('kp_quality', np.float32, uint8_nan),
            'num_val_trip': ('num_val', np.float32, None)
        }

        skip_fields = [
            'f_f', 'f_v', 'f_oa', 'f_sa', 'f_tel', 'f_ref', 'abs_line_number'
        ]

        data, metadata = read_nc(filename, generic, to_xarray,
                                 skip_fields, gen_fields_lut)

        return data, metadata

    def _merge(self, data):
        """
        Merge data.

        Parameters
        ----------
        data : list
            List of array.

        Returns
        -------
        data : numpy.ndarray
            Data.
        """
        if isinstance(data[0], tuple):
            data, metadata = zip(*data)
            if isinstance(data[0], xr.Dataset):
                data = xr.concat(data,
                                 dim="obs",
                                 combine_attrs="drop_conflicts")
            else:
                data = np.hstack(data)
            data = (data, metadata)
        else:
            data = np.hstack(data)

        return data

class AscatL1bNcFileGeneric(AscatL1bNcFile):
    """
    The same as AscatL1bNcFile but with generic=True by default.
    """
    def _read(self, filename, generic=True, to_xarray=False, **kwargs):
        return super()._read(filename, generic=generic, to_xarray=to_xarray, **kwargs)


class AscatL2NcFile(AscatFile):
    """
    Read ASCAT Level 2 file in NetCDF format.
    """

    def __init__(self, filename, **kwargs):
        """
        Initialize AscatL2NcFile.

        Parameters
        ----------
        filename : str
            Filename.
        """
        super().__init__(filename, **kwargs)
        for i, fname in enumerate(self.filenames):
            if os.path.splitext(fname)[1] == '.gz':
                self.filenames[i] = tmp_unzip(fname)

    def _read(self, filename, generic=False, to_xarray=False):
        """
        Read one ASCAT Level 2 NetCDF4 file.

        Parameters
        ----------
        generic : bool, optional
            'True' reading and converting into generic format or
            'False' reading original field names (default: False).
        to_xarray : bool, optional
            'True' return data as xarray.Dataset
            'False' return data as numpy.ndarray (default: False).

        Returns
        -------
        ds : dict, xarray.Dataset
            ASCAT Level 2 data.
        """
        gen_fields_lut = {
            'longitude': ('lon', np.float32, None),
            'latitude': ('lat', np.float32, None),
            'utc_line_nodes': ('time', np.float32, None),
            'inc_angle_trip': ('inc', np.float32, float32_nan),
            'azi_angle_trip': ('azi', np.float32, float32_nan),
            'sigma0_trip': ('sig', np.float32, float32_nan),
            'kp': ('kp', np.float32, float32_nan),
            'soil_moisture': ('sm', np.float32, float32_nan),
            'soil_moisture_error': ('sm_noise', np.float32, float32_nan),
            'sigma40': ('sig40', np.float32, float32_nan),
            'sigma40_error': ('sig40_noise', np.float32, float32_nan),
            'slope40': ('slope40', np.float32, float32_nan),
            'slope40_error': ('slope40_noise', np.float32, float32_nan),
            'soil_moisture_sensitivity': ('sm_sens', np.float32, float32_nan),
            'dry_backscatter': ('dry_sig40', np.float32, float32_nan),
            'wet_backscatter': ('wet_sig40', np.float32, float32_nan),
            'mean_soil_moisture': ('sm_mean', np.float32, float32_nan),
            'proc_flag1': ('corr_flag', np.uint8, None),
            'proc_flag2': ('proc_flag', np.uint8, None),
            'aggregated_quality_flag': ('agg_flag', np.uint8, None),
            'snow_cover_probability': ('snow_prob', np.uint8, None),
            'frozen_soil_probability': ('frozen_prob', np.uint8, None),
            'wetland_flag': ('wetland', np.uint8, None),
            'topography_flag': ('topo', np.uint8, None)
        }

        skip_fields = ['abs_line_number']

        data, metadata = read_nc(filename, generic, to_xarray,
                                 skip_fields, gen_fields_lut)

        return data, metadata

    def _merge(self, data):
        """
        Merge data.

        Parameters
        ----------
        data : list
            List of array.

        Returns
        -------
        data : numpy.ndarray
            Data.
        """
        if isinstance(data[0], tuple):
            data, metadata = zip(*data)
            if isinstance(data[0], xr.Dataset):
                data = xr.concat(data,
                                 dim="obs",
                                 combine_attrs="drop_conflicts")
            else:
                data = np.hstack(data)
            data = (data, metadata)
        else:
            data = np.hstack(data)

        return data

class AscatL2NcFileGeneric(AscatL2NcFile):
    """
    The same as AscatL1bNcFile but with generic=True by default.
    """
    def _read(self, filename, generic=True, to_xarray=False, **kwargs):
        return super()._read(filename, generic=generic, to_xarray=to_xarray, **kwargs)


class AscatSsmNcSwathFile(AscatFile):
    """
    Class reading ASCAT Surface Soil Moisture Netcdf swath file.
    """

    def _read(self, filename, mask_and_scale=None, sel_dt=None):
        """
        Read/load data from NetCDF file.

        Parameters
        ----------
        mask_and_scale : boolean, optional
            Mask and scale data using _FillValue.

        Returns
        -------
        data : xarray.Dataset
            Data.
        """
        with xr.open_dataset(self.filename,
                             mask_and_scale=mask_and_scale) as ds:
            data = ds.load()

        if sel_dt is not None:
            data = data.where((data.time >= sel_dt[0])
                              & (data.time <= sel_dt[1]),
                              drop=True)

        return data


class AscatSsmNcSwathFileList(ChronFiles):
    """
    Class reading ASCAT Surface Soil Moisture Netcdf swath file list.
    """

    def __init__(self,
                 path,
                 filename_template=None,
                 subfolder_template=None,
                 sat="?",
                 cls_kwargs=None):
        """
        Initialize object.

        Parameters
        ----------
        path : str
            Root path to data.
        filename_template : str, optional
            Filename template (default: "W_IT-HSAF-ROME,SAT,SSM-ASCAT-METOP{sat}-6.25-H???_C_LIIB_{date}_*_*____.nc")
        subfolder_template : str, optional
            Subfolder template (default: /path/metop_{sat}/{year}/)
        sat : str, optional
            Satellite Metop-A: "A", Metop-B: "B", Metop-C: "C" or all: "?"
            (default: "?")
        cls_kwargs : dict, optional
            Keyword arguments passed to file IO class (default: None).
        """
        if filename_template is None:
            filename_template = "W_IT-HSAF-ROME,SAT,SSM-ASCAT-METOP{sat}-6.25-H???_C_LIIB_{date}_*_*____.nc"

        if subfolder_template is None:
            subfolder_template = {"satellite": "metop_{sat}", "date": "{year}"}

        if cls_kwargs is None:
            cls_kwargs = {}

        self.sat = sat

        super().__init__(path,
                         AscatSsmNcSwathFile,
                         filename_template,
                         sf_templ=subfolder_template,
                         cls_kwargs=cls_kwargs)

    def _fmt(self, timestamp):
        """
        Definition of filename and subfolder format.

        Parameters
        ----------
        timestamp : datetime
            Time stamp.

        Returns
        -------
        fn_fmt : dict
            Filename format.
        sf_fmt : dict
            Subfolder format.
        """
        fn_read_fmt = {
            "date": timestamp.strftime("%Y%m%d%H%M%S"),
            "sat": self.sat.upper()
        }
        fn_write_fmt = None

        sf_read_fmt = {
            "satellite": {
                "sat": self.sat
            },
            "date": {
                "year": timestamp.strftime("%Y"),
                "month": timestamp.strftime("%m"),
                "day": timestamp.strftime("%d")
            }
        }
        sf_write_fmt = None

        return fn_read_fmt, sf_read_fmt, fn_write_fmt, sf_write_fmt

    def _parse_date(self, filename, start=53, end=64):
        """
        Parse date from filename.

        Parameters
        ----------
        filename : str
            Filename.
        start : int, optional
            Start position of date field (default: 58).
        start : int, optional
            End position  of date field (default: 72).

        Returns
        -------
        date : datetime
            Parsed date.
        """
        return datetime.strptime(
            os.path.basename(filename)[start:end], "%Y%m%d%H%M%S")

    def search_date(self, timestamp, **kwargs):
        """
        Search date.

        Parameters
        ----------
        timestamp : datetime
            Date.

        Returns
        -------
        filenames : list
            Filenames.
        """
        return super().search_date(timestamp, date_str="%Y%m%d%H*", **kwargs)

    def iter_daterange(self, start_date, end_date):
        """
        Generator returning filenames between start and end date.

        Parameters
        ----------
        start_date : datetime
            Start date.
        end_date : datetime
            End date.

        Yields
        ------
        filename : str
            Filename.
        """
        for single_date in daterange(start_date, end_date):
            filenames = self.search_date(single_date)
            for filename in filenames:
                yield filename

    def read_date(self, timestamp):
        """
        Read data for given timestamp.

        Parameters
        ----------
        timestamp : datetime
            Date.

        Returns
        -------
        data : xarray.Dataset
            Data.
        """
        filenames = self.search_date(timestamp)

        if len(filenames) > 1:
            raise RuntimeError(
                f"Multiple files found for timestamp {timestamp}")
        elif len(filenames) == 0:
            print(f"No file found for timestamp {timestamp}")
            data = None
        else:
            self._open(filenames[0])
            data = self.fid.read()

        return data

    def read_period(self,
                    start_dt,
                    end_dt,
                    delta_dt=timedelta(hours=1),
                    buffer_dt=timedelta(hours=1),
                    **kwargs):
        """
        Read data for given interval.

        Parameters
        ----------
        start_dt : datetime
            Start datetime.
        end_dt : datetime
            End datetime.
        delta_dt : timedelta, optional
            Time delta used to jump through search date.
        buffer_dt : timedelta, optional
            Search buffer used to find files which could possibly contain
            data but would be left out because of dt_start.

        Returns
        -------
        data : dict, numpy.ndarray
            Data stored in file.
        """
        filenames = self.search_period(start_dt - buffer_dt,
                                       end_dt + buffer_dt, delta_dt)

        merged_data = []

        sel_dt = (np.datetime64(start_dt - timedelta(minutes=15)),
                  np.datetime64(end_dt + timedelta(minutes=15)))

        for filename in filenames:
            self._open(filename)

            try:
                data = self.fid.read(sel_dt=sel_dt)

                if data is not None:
                    merged_data.append(data)
            except:
                print(f"Error reading: {self.fid.filename}")

        if merged_data:
            merged_data = xr.concat(merged_data,
                                    dim="obs",
                                    combine_attrs="drop_conflicts")
        else:
            merged_data = None

        return merged_data



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Download interface.
"""

import sys
import datetime
import argparse
import configparser

from ascat.download.connectors import HsafConnector
from ascat.download.connectors import EumConnector


def parse_date(s):
    """
    Convert date string into datetime timestamp.

    Parameters
    ----------
    s : str
        Date string.

    Returns
    -------
    data : datetime
        Timestamp.
    """
    return datetime.datetime.strptime(s, '%Y%m%d')


def hsaf_download(credentials,
                  remote_path,
                  local_path,
                  start_date,
                  end_date,
                  limit=None):
    """
    Function to start H SAF download.

    Parameters
    ----------
    credentials : dict
        Dictionary of needed authentication parameters ('user', 'password').
    remote_path : string
        Remote directory, where found datasets are stored.
    local_path : string
        Local directory, where found datasets are stored.
    start_date : datetime
        Start date of date range interval.
    end_date : datetime
        End date of date range interval.
    limit : int, optional
        Filter used to limit the returned results (default: 1).
    """
    connector = HsafConnector()
    connector.connect(credentials)
    connector.download(remote_path, local_path, start_date, end_date, limit)
    connector.close()


def eumetsat_download(credentials,
                      product,
                      local_path,
                      start_date,
                      end_date,
                      max_workers=1,
                      coords=None,
                      limit=None):
    """
    Function to start EUMETSAT download.

    Parameters
    ----------
    credentials : dict
        Dictionary of needed authentication parameters ('user', 'password').
    remote_path : string
        Remote directory, where found datasets are stored.
    local_path : string
        Local directory, where found datasets are stored.
    start_date : datetime
        Start date of date range interval.
    end_date : datetime
        End date of date range interval.
    max_workers : int, optional
        Number of parallel downloads (default: 1).
    coords : list of float, optional
        A custom polygon using EPSG:4326 decimal degrees (default: None).
    limit : int, optional
        Filter used to limit the returned results (default: None).
    """
    connector = EumConnector()
    connector.connect(credentials)
    connector.download(product, local_path, start_date, end_date, max_workers,
                       coords, limit)


def parse_args_hsaf_download(args):
    """
    Parse command line arguments.

    Parameters
    ----------
    args : list
        Command line arguments.

    Returns
    -------
    args : list
        Parsed arguments.
    parser : ArgumentParser
        Argument Parser object.
    """
    parser = argparse.ArgumentParser(
        description='H SAF download command line interface.',
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-cf',
                        '--credential_file',
                        help='File where credentials are stored')

    parser.add_argument('-r', '--remote_path', help='FTP remote path')

    parser.add_argument('-o', '--output_dir', help='Directory to write output')

    parser.add_argument('-from',
                        '--start_date',
                        type=parse_date,
                        help='start date in YYYYMMDD format')

    parser.add_argument('-to',
                        '--end_date',
                        type=parse_date,
                        help='end date in YYYYMMDD format')

    parser.add_argument(
        '-co',
        '--coords',
        help='A custom polygon using EPSG:4326 decimal degrees')

    parser.add_argument('-l', '--limit', help='Filter number of results')

    return parser.parse_args(args), parser


def parse_args_eumetsat_download(args):
    """
    Parse command line arguments.

    Parameters
    ----------
    args : list
        Command line arguments.

    Returns
    -------
    args : list
        Parsed arguments.
    parser : ArgumentParser
        Argument Parser object.
    """
    parser = argparse.ArgumentParser(
        description='EUMETSAT download command line interface.',
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-cf',
                        '--credential_file',
                        help='File where credentials are stored')

    parser.add_argument('-p', '--product', help='Name of product')

    parser.add_argument('-o', '--output_dir', help='Directory to write output')

    parser.add_argument('-from',
                        '--start_date',
                        type=parse_date,
                        help='start date in YYYYMMDD format')

    parser.add_argument('-to',
                        '--end_date',
                        type=parse_date,
                        help='end date in YYYYMMDD format')

    parser.add_argument('-mw',
                        '--max_workers',
                        type=int,
                        help='Number of parallel downloads')

    parser.add_argument(
        '-co',
        '--coords',
        help='A custom polygon using EPSG:4326 decimal degrees')

    parser.add_argument('-l', '--limit', help='Filter number of results')

    return parser.parse_args(args), parser


def hsaf_main(cli_args):
    """
    Command line interface routine.

    Parameters
    ----------
    cli_args : list
        Command line arguments.
    """
    args, parser = parse_args_hsaf_download(cli_args)

    credentials = configparser.ConfigParser()
    credentials.read(args.credential_file)

    hsaf_download(credentials['hsaf'], args.remote_path, args.output_dir,
                  args.start_date, args.end_date, args.limit)


def eumetsat_main(cli_args):
    """
    Command line interface routine.

    Parameters
    ----------
    cli_args : list
        Command line arguments.
    """
    args, parser = parse_args_eumetsat_download(cli_args)

    credentials = configparser.ConfigParser()
    credentials.read(args.credential_file)

    eumetsat_download(credentials['eumetsat'], args.product, args.output_dir,
                      args.start_date, args.end_date, args.max_workers,
                      args.coords, args.limit)


def run_hsaf_download():
    """
    Run command line interface for H SAF download.
    """
    hsaf_main(sys.argv[1:])


def run_eumetsat_download():
    """
    Run command line interface for EUMETSAT download.
    """
    eumetsat_main(sys.argv[1:])



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from functools import partial
from pathlib import Path

import xarray as xr
import numpy as np

import ascat.accessors
from ascat.grids import GridRegistry

from ascat.file_handling import Filenames
from ascat.utils import get_grid_gpis
from ascat.utils import append_to_netcdf
import warnings


class RaggedArrayTs(Filenames):
    """
    Class to read and merge ragged array cell files.
    """
    def _read(
        self,
        filename,
        location_id=None,
        lookup_vector=None,
        date_range=None,
        preprocessor=None,
        **xarray_kwargs
    ):
        """
        Open one Ragged Array file as an xarray.Dataset and preprocess it if necessary.

        Parameters
        ----------
        filename : str
            File to read.
        location_id : int or list of int
            Location id. Only used for selecting points from contiguous ragged arrays.
            Not used for indexed ragged arrays or point arrays.
        lookup_vector : np.ndarray
            Lookup vector for faster selection.
        date_range : tuple of np.datetime64
            Tuple of (start, end) dates.
        preprocessor : callable, optional
            Function to preprocess the dataset.
        xarray_kwargs : dict
            Additional keyword arguments passed to xarray.open_dataset.

        Returns
        -------
        ds : xarray.Dataset
            Dataset.
        """
        if ds := self.cache.get(filename):
            pass
        else:
            ds = xr.open_dataset(filename, engine = "h5netcdf", **xarray_kwargs)
            self.cache[filename] = ds

        if preprocessor:
            ds = preprocessor(ds)
        ds = self._ensure_obs(ds)

        if ds.cf_geom.array_type == "contiguous":
            ds = self._trim_to_gpis(ds, gpis=location_id, lookup_vector=lookup_vector)

        elif ds.cf_geom.array_type == "indexed":
            ds.time.load()
            # we need to make sure the time variable is in memory before converting to
            # a point array, since reordering this as a dask array will explode the process
            # graph later on
            ds = self._ensure_point(ds)
            if date_range is not None:
                ds = self._trim_var_range(ds, "time", *date_range)

        return ds


    def read(self,
             date_range=None,
             location_id=None,
             lookup_vector=None,
             preprocessor=None,
             return_format=None,
             parallel=False,
             **kwargs):
        """
        Read data from Ragged Array Cell files.

        Parameters
        ----------
        date_range : tuple of np.datetime64
            Tuple of (start, end) dates.
        location_id : list of int
            List of timeseries IDs to read.
        lookup_vector : np.ndarray
            Lookup vector.
        preprocessor : callable, optional
            Function to preprocess the dataset.
        return_format : str, optional
            CF discrete geometry format to return data as. Can be "point", "indexed", or "contiguous".
        parallel : bool, optional
            Whether or not to read/preprocess in parallel. Default is False.
        **kwargs : dict
        """
        ds, closers = super().read(date_range=date_range,
                                   location_id=location_id,
                                   lookup_vector=lookup_vector,
                                   preprocessor=preprocessor,
                                   closer_attr="_close",
                                   parallel=parallel,
                                   **kwargs)

        if ds is not None:
            ds.set_close(partial(super()._multi_file_closer, closers))

            if ds.cf_geom.array_type == "contiguous" and date_range is not None:
                ds = self._dim_safe_rechunk(ds)
                ds = self._trim_var_range(ds, "time", *date_range)
                ds = self._dim_safe_rechunk(ds)

            if ds.cf_geom.array_type != "contiguous":
                ds = self._trim_to_gpis(ds.chunk({"obs": 1000000}), gpis=location_id, lookup_vector=lookup_vector)

            if return_format is not None:
                if return_format == "point":
                    return ds.cf_geom.to_point_array()

                elif return_format == "indexed":
                    return ds.cf_geom.to_indexed_ragged()

                elif return_format == "contiguous":
                    return ds.cf_geom.to_contiguous_ragged()
                else:
                    raise ValueError("return_format must be 'point', 'indexed', "
                                     "'contiguous', or None (default, returns as read)")
            return ds


    @staticmethod
    def _indexed_or_contiguous(ds):
        if "locationIndex" in ds:
            return "indexed"
        return "contiguous"

    @staticmethod
    def _array_type(ds):
        """
        Determine the type of ragged array.
        """
        if "locationIndex" in ds:
            return "indexed"
        if "row_size" in ds:
            return "contiguous"

    @staticmethod
    def _trim_var_range(ds, var_name, var_min, var_max, end_inclusive=False):
        if end_inclusive:
            mask = (ds[var_name] >= var_min) & (ds[var_name] <= var_max)
        else:
            mask = (ds[var_name] >= var_min) & (ds[var_name] < var_max)

        mask = mask.compute()
        if ds.cf_geom.array_type == "contiguous" and "locations" in ds.dims:
            group_indices = np.repeat(np.arange(ds.row_size.size), ds.row_size)
            kept_counts = np.bincount(group_indices,
                                      weights=mask.astype(int),
                                      minlength=ds.row_size.size).astype(np.int32)
            ds = ds.sel(obs=mask)
            ds["row_size"].values = kept_counts
            return ds
        else:
            return ds.sel(obs=mask)


    @staticmethod
    def _ensure_obs(ds):
        ds = ds.cf_geom.set_sample_dimension("obs")
        return ds

    @staticmethod
    def _ensure_point(ds):
        return ds.cf_geom.to_point_array()

    @staticmethod
    def _ensure_indexed(ds):
        """
        Convert a contiguous dataset to indexed dataset,
        if necessary. Indexed datasets pass through.

        Ragged array type is determined by the presence of
        either a "row_size" or "locationIndex" variable
        (for contiguous and indexed arrays, respectively).

        Parameters
        ----------
        ds : xarray.Dataset
            Dataset in indexed or contiguous ragged array format.

        Returns
        -------
        xarray.Dataset
            Dataset in indexed ragged array format.
        """
        if ds is not None:
            return ds.cf_geom.to_indexed_ragged()


    @staticmethod
    def _ensure_contiguous(ds):
        """
        Convert an indexed dataset to contiguous dataset,
        if necessary. Contiguous datasets pass through.

        Ragged array type is determined by the presence of
        either a "row_size" or "locationIndex" variable
        (for contiguous and indexed arrays, respectively).

        Parameters
        ----------
        ds : xarray.Dataset, Path
            Dataset in indexed or contiguous ragged array format.

        Returns
        -------
        xarray.Dataset
            Dataset in contiguous ragged array format.
        """
        if ds is not None:
            return ds.cf_geom.to_contiguous_ragged(sort_vars=["time"])

    @staticmethod
    def _only_locations(ds):
        """Return a dataset with only the variables that aren't in the obs-dimension.

        Parameters
        ----------
        ds : xarray.Dataset
            Dataset.

        Returns
        -------
        xarray.Dataset
            Dataset with only the locations-dimensional variables.
        """
        return ds[[
            var
            for var in ds.variables
            if "obs" not in ds[var].dims
            and var not in ["row_size", "locationIndex"]
        ]]

    @staticmethod
    def _dim_safe_rechunk(data):
        if "locations" in data.dims:
            return data.chunk({"obs": 1000000, "locations": -1})
        return data.chunk({"obs": 1000000})


    def _merge(self, data):
        """
        Merge datasets with potentially different locations dimensions.

        Parameters
        ----------
        data : list of xarray.Dataset
            Datasets to merge.

        Returns
        -------
        xarray.Dataset
            Merged dataset.
        """
        if data == []:
            return None

        if data[0].cf_geom.array_type == "indexed":
            return self._merge_indexed(data)
        elif data[0].cf_geom.array_type == "point":
            return self._merge_point(data)
        elif data[0].cf_geom.array_type == "contiguous":
            return self._merge_contiguous(data)
        else:
            raise ValueError("Array type must be 'contiguous', 'indexed', or 'point'")


    def _merge_indexed(self, data):
        preprocessed_point_arrays_to_merge = [(self._preprocess(ds)
                                               .chunk({"obs":-1})
                                               .cf_geom.to_point_array())
                                              for ds in data if ds is not None]
        merged_point_ds = xr.concat(
            preprocessed_point_arrays_to_merge,
            dim="obs",
            data_vars="minimal",
            compat="equals",
            combine_attrs="drop_conflicts",
        )
        merged_indexed_ds = merged_point_ds.cf_geom.to_indexed_ragged()
        return merged_indexed_ds

    def _merge_point(self, data):
        merged_ds = xr.concat(
            [self._preprocess(ds).chunk({"obs":-1})
                for ds in data if ds is not None],
            dim="obs",
            combine_attrs="drop_conflicts",
        )
        return merged_ds

    @staticmethod
    def _only_obs_vars(ds):
        return ds[[var for var in ds.variables if "obs" in ds[var].dims]]

    def _merge_contiguous(self, data):
        preprocessed = [self._preprocess(ds).chunk({"obs": -1})
                        for ds in data if ds is not None]
        obs_dim_ds_to_concat = [self._only_obs_vars(ds) for ds in preprocessed]
        non_obs_dim_ds_to_concat = [ds.drop_dims("obs") for ds in preprocessed]
        obs_dim_ds = xr.concat(obs_dim_ds_to_concat, dim="obs")
        # any variables without dims will be in non_obs_dim_ds, we handle them
        # by assuming they should be equal on all merged data - not sure if this
        # is the right choice. We could also override.
        non_obs_dim_ds = xr.concat(non_obs_dim_ds_to_concat,
                                   dim="locations",
                                   data_vars="minimal",
                                   compat="equals")
        merged_ds = xr.merge([obs_dim_ds, non_obs_dim_ds])
        merged_ds = self._dim_safe_rechunk(merged_ds)
        return merged_ds

    def _preprocess(self, ds):
        """Pre-processing to be done on a component dataset so it can be merged with others.

        Nothing here at the moment.

        Parameters
        ----------
        ds : xarray.Dataset
            Dataset.

        Returns
        -------
        xarray.Dataset
            Dataset with pre-processing applied.
        """
        return ds

    def _trim_to_gpis(self, ds, gpis=None, lookup_vector=None):
        """Trim a dataset to only the gpis in the given list.
        If any gpis are passed which are not in the dataset, they are ignored.

        Parameters
        ----------
        ds : xarray.Dataset
            Dataset.
        gpis : list or list-like
            List of gpis to keep. One of gpis or lookup_vector
            must be provided.
        lookup_vector : np.ndarray
            Lookup vector from gpi numbers to bools for inclusion. One of gpis or lookup_vector
            must be provided.

        Returns
        -------
        xarray.Dataset
            Dataset with only the gpis in the list.
        """
        if ds is None:
            return

        if (gpis is None or len(gpis) == 0) and (lookup_vector is None or len(lookup_vector)==0):
            return ds

        return ds.cf_geom.sel_instances(gpis, lookup_vector)

    def _write(self,
               data,
               filename,
               ra_type="indexed",
               mode="w",
               postprocessor=None,
               **kwargs):
        """
        Write data to a netCDF file.

        Parameters
        ----------
        filename : str
            Output filename.
        ra_type : str, optional
            Type of ragged array to write. Default is "contiguous".
        **kwargs : dict
            Additional keyword arguments passed to xarray.to_netcdf().
        """
        if ra_type == "contiguous":
            data = self._ensure_contiguous(data)
        if ra_type == "indexed":
            data = self._ensure_indexed(data)
        if ra_type == "point":
            data = self._ensure_point(data)

        if postprocessor is not None:
            data = postprocessor(data)

        data.encoding["unlimited_dims"] = ["obs"]

        if mode == "a" and ra_type in ["indexed", "point"]:
            if Path(filename).exists():
                append_to_netcdf(filename, data, unlimited_dim="obs")
                data.close()
                return

        import warnings
        with warnings.catch_warnings():
            # NetCDF will sometimes warn us about endianness for some reason and idk why
            warnings.filterwarnings("ignore",
                                    message="endian-ness of dtype and endian kwarg do not match, using endian kwarg",
                                    category=UserWarning)
            data.to_netcdf(filename, **kwargs)
        data.close()


class OrthoMultiTimeseriesCell(Filenames):
    """
    Class to read and merge orthomulti cell files.
    """
    def _read(self, filename, preprocessor=None, **xarray_kwargs):
        """
        Open one OrthoMulti file as an xarray.Dataset and preprocess it if necessary.

        Parameters
        ----------
        filename : str
            File to read.
        preprocessor : callable, optional
            Function to preprocess the dataset.
        xarray_kwargs : dict
            Additional keyword arguments passed to xarray.open_dataset.

        Returns
        -------
        ds : xarray.Dataset
            Dataset.
        """
        ds = xr.open_dataset(filename, **xarray_kwargs)
        if preprocessor:
            ds = preprocessor(ds)
        return ds

    def read(self,
             date_range=None,
             location_id=None,
             lookup_vector=None,
             preprocessor=None,
             parallel=False,
             **kwargs):
        """
        Read data from OrthoMulti Cell files.

        Parameters
        ----------
        date_range : tuple of np.datetime64
            Tuple of (start, end) dates.
        location_id : list of int
            List of timeseries IDs to read.
        lookup_vector : np.ndarray
            Lookup vector.
        preprocessor : callable, optional
            Function to preprocess the dataset.
        parallel : bool, optional
            Whether or not to read/preprocess in parallel. Default is False.
        """
        ds = super().read(preprocessor=preprocessor, **kwargs)
        if date_range is not None:
            ds = ds.sel(time=slice(*date_range))
        ds = self._trim_to_gpis(ds, gpis=location_id, lookup_vector=lookup_vector)

        return ds

    def _merge(self, data):
        """
        Merge datasets with different location and/or time dimensions.

        Parameters
        ----------
        data : list of xarray.Dataset
            Datasets to merge.

        Returns
        -------
        xarray.Dataset
            Merged dataset.
        """
        if data == []:
            return None

        # ensures that time dimensions are aligned in case they are not
        data = xr.align(*data, join="outer")

        data = [ds.chunk(-1).set_index(locations="location_id")
                for ds in data if ds is not None]
        merged_ds = xr.concat(
            data,
            dim="locations",
            combine_attrs="drop_conflicts",
        ).rename_vars({"locations": "location_id"})
        merged_ds["location_id"].attrs["cf_role"] = "timeseries_id"

        return merged_ds

    @staticmethod
    def _trim_to_gpis(ds, gpis=None, lookup_vector=None):
        """Trim a dataset to only the gpis in the given list.
        If any gpis are passed which are not in the dataset, they are ignored.

        Parameters
        ----------
        ds : xarray.Dataset
            Dataset.
        gpis : list or list-like
            List of gpis to keep.
        lookup_vector : np.ndarray
            Lookup vector from gpi numbers to bools for inclusion.

        Returns
        -------
        xarray.Dataset
            Dataset with only the gpis in the list.
        """
        if ds is None:
            return

        if (gpis is None or len(gpis) == 0) and (lookup_vector is None or len(lookup_vector)==0):
            return ds

        return ds.cf_geom.sel_instances(gpis, lookup_vector)


class CellGridFiles():

    def __init__(
        self,
        root_path,
        file_class,
        grid,
        fn_format="{cell:04d}.nc",
        sf_format=None,
        preprocessor=None,
    ):
        """
        Initialize cell grid files.

        Parameters
        ----------
        root_path : str or Path
            Root path where the cell files are stored.
        file_class : class
            Class to use for file handling (e.g., RaggedArrayTs or OrthoMultiTimeseriesCell).
        grid : str or Grid
            Grid name or object defining the spatial grid structure.
        fn_format : str, optional
            Format string for cell file names, default is "{cell:04d}.nc".
        sf_format : str, optional
            Format string for subdirectories, if any.
        preprocessor : callable, optional
            Function to preprocess datasets when reading.
        """
        self.root_path = Path(root_path)
        self.file_class = file_class
        self.grid = grid
        self.fn_format = fn_format
        self.sf_format = sf_format
        self._preprocessor = preprocessor
        self._active_reader = None


    @classmethod
    def from_product_id(cls, root_path, product_id, **kwargs):
        from ascat.product_info import cell_io_catalog
        product_id = product_id.upper()
        if product_id in cell_io_catalog:
            product_class = cell_io_catalog[product_id]
            return cls.from_product_class(root_path, product_class, **kwargs)
        error_str = f"Product {product_id} not recognized. Valid products are"
        error_str += f" {', '.join(cell_io_catalog.keys())}."
        raise ValueError(error_str)

    @classmethod
    def from_product_class(cls, root_path, product_class, **kwargs):
        grid_name = product_class.grid_name
        init_options = {
            "root_path": root_path,
            "file_class": product_class.file_class,
            "grid": GridRegistry().get(grid_name),
            "fn_format": product_class.fn_format,
            "preprocessor": product_class.preprocessor,
            **kwargs
        }
        init_options = {**init_options}
        return cls(**init_options)

    def fn_search(self, cell):
        # get the paths to files matching a cell if the files exist
        filename = self.fn_format.format(cell)
        if self.sf_format is not None:
            subfolder = self.sf_format
            files = list(self.root_path.glob(subfolder + "/" + filename))
            return files
        else:
            return list(self.root_path.glob("**/" + filename))

    def convert_to_contiguous(self, out_dir, print_progress=True, **kwargs):
        """
        Convert all files in the collection to contiguous format and write to disk.

        Parameters
        ----------
        out_dir : str
            Output directory.
        print_progress : bool, optional
            Whether to print progress messages to console. Default is True.
        kwargs : dict
            Keyword arguments passed to the reprocess method.
        """
        self.reprocess(out_dir, lambda ds: ds, ra_type="contiguous", print_progress=True, **kwargs)

    def reprocess(self, out_dir, func, parallel=True, **kwargs):
        """
        Use Filenames.reprocess to apply a function to all files in the collection and
        save the results to `out_dir`.

        Parameters
        ----------
        out_dir : str
            Output directory.
        func : callable
            Function to apply to each file.
        parallel : bool, optional
            Whether to process files in parallel. Default is True.
        kwargs : dict
            Keyword arguments passed to func.
        """

        # get list of every filename in the collection
        filenames = self.spatial_search()
        files = self.file_class(filenames)
        files.reprocess(out_dir,
                        func,
                        parallel=parallel,
                        read_kwargs={"preprocessor": self._preprocessor},
                        **kwargs)


    def spatial_search(
            self,
            cell=None,
            location_id=None,
            coords=None,
            bbox=None,
            geom=None,
    ):
        """
        Search files for cells matching a spatial criterion. All args are declared
        as optional; but one and only one should be passed.

        Parameters
        ----------
        cell : int or list of int
            Grid cell number to read.
        location_id : int or list of int
            Location id.
        coords : tuple of numeric or tuple of iterable of numeric
            Tuple of (lon, lat) coordinates.
        bbox : tuple
            Tuple of (latmin, latmax, lonmin, lonmax) coordinates.
        geom : shapely.geometry
            Geometry object.

        Returns
        -------
        filenames : list of str
            Filenames.
        """
        if cell is not None:
            # guarantee cell is a list
            matched_cells = cell
            if not isinstance(matched_cells, list):
                matched_cells = [matched_cells]
        elif location_id is not None:
            # guarantee location_id is a list
            if not isinstance(location_id, list):
                location_id = [location_id]
            matched_cells = self._cells_for_location_id(location_id)
        elif coords is not None:
            matched_cells = self._cells_for_coords(coords)
        elif bbox is not None:
            matched_cells = self._cells_for_bbox(bbox)
        elif geom is not None:
            matched_cells = self._cells_for_geom(geom)
        else:
            matched_cells = self.grid.arrcell

        matched_cells = np.unique(matched_cells)

        filenames = []
        for cell in matched_cells:
            filenames += self.fn_search(cell)

        return filenames

    def _cells_for_location_id(self, location_id):
        """
        Get cells for location_id.

        Parameters
        ----------
        location_id : int
            Location id.

        Returns
        -------
        cells : list of int
            Cells.
        """
        cells = self.grid.gpi2cell(location_id)
        return cells

    def _cells_for_coords(self, coords):
        """
        Get cells for coordinates.

        Parameters
        ----------
        coords : tuple
            Coordinates (lon, lat)

        Returns
        -------
        cells : list of int
            Cells.
        """
        gpis = get_grid_gpis(self.grid, coords=coords)
        cells = self._cells_for_location_id(gpis)
        return cells

    def _cells_for_bbox(self, bbox):
        """
        Get cells for bounding box.

        Parameters
        ----------
        bbox : tuple
            Bounding box.

        Returns
        -------
        cells : list of int
            Cells.
        """
        gpis = get_grid_gpis(self.grid, bbox=bbox)
        cells = self._cells_for_location_id(gpis)
        return cells

    def _cells_for_geom(self, geom):
        """
        Get cells for geometry.

        Parameters
        ----------
        geom : shapely.geometry
            Geometry object.

        Returns
        -------
        cells : list of int
            Cells.
        """
        gpis = get_grid_gpis(self.grid, geom=geom)
        cells = self._cells_for_location_id(gpis)
        return cells

    def _fn(self, cell):
        return self.root_path / self.fn_format.format(cell)

    def read(
            self,
            cell=None,
            location_id=None,
            coords=None,
            bbox=None,
            geom=None,
            max_coord_dist=np.inf,
            date_range=None,
            **kwargs,
    ):
        """
        Read data matching a spatial and temporal criterion.

        Parameters
        ----------
        cell : int or list of int
            Grid cell number to read.
        location_id : int or list of int
            Location id.
        coords : tuple of numeric or tuple of iterable of numeric
            Tuple of (lon, lat) coordinates. lon and lat could each be numpy arrays in
            order to read multiple coordinates. For each coordinate the nearest grid point
            within `max_coord_dist` (in spherical cartesian coordinates) will be selected.

            Note that if any passed coordinates share the same nearest grid point, that grid
            point will only be represented once in the output dataset.
        bbox : tuple
            Tuple of (latmin, latmax, lonmin, lonmax) coordinates.
        geom : shapely.geometry
            Geometry object.
        max_coord_dist : float
            The maximum distance a coordinate's nearest grid point can be from it to be
            selected (in spherical cartesian coordinates). Default is np.inf.
        date_range : tuple of np.datetime64
            Tuple of (start, end) dates.

        Returns
        -------
        xarray.Dataset
            Filtered and merged data for the specified spatiotemporal region.

        """
        filenames = self.spatial_search(
            cell=cell,
            location_id=location_id,
            coords=coords,
            bbox=bbox,
            geom=geom,
        )
        if ((self._active_reader is None)
            or not
            all(filename in self._active_reader.cache for filename in filenames)):
            self._active_reader = self.file_class(filenames)

        if all(criterion is None for criterion in [cell, location_id, coords, bbox, geom]):
            valid_gpis = None
            lookup_vector = None
        elif cell is not None:
            valid_gpis = None
            lookup_vector = None
        else:
            valid_gpis, lookup_vector = get_grid_gpis(
                self.grid,
                cell,
                location_id,
                coords,
                bbox,
                geom,
                max_coord_dist,
                return_lookup=True
            )

        out_ds = self._active_reader.read(date_range=date_range,
                                          location_id=valid_gpis,
                                          lookup_vector=lookup_vector,
                                          preprocessor=self._preprocessor,
                                          **kwargs)

        if out_ds is None:
            warning_str = ("No data found for specified criteria, returning None:\n"
                           f"cell={cell}, location_id={location_id}, coords={coords}, bbox={bbox}, geom={geom},\n"
                           f"max_coord_dist={max_coord_dist}, date_range={date_range}")
            warnings.warn(warning_str, UserWarning, 2)
            return out_ds
        return out_ds



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os

from datetime import timedelta
from gzip import GzipFile
from tempfile import NamedTemporaryFile

import numpy as np
import xarray as xr

import netCDF4

from shapely.geometry import Point

int8_nan = np.iinfo(np.int8).min
uint8_nan = np.iinfo(np.uint8).max
int16_nan = np.iinfo(np.int16).min
uint16_nan = np.iinfo(np.uint16).max
int32_nan = np.iinfo(np.int32).min
uint32_nan = np.iinfo(np.uint32).max
int64_nan = np.iinfo(np.int64).min
uint64_nan = np.iinfo(np.uint64).max
float32_nan = -999999.
float64_nan = -999999.

dtype_to_nan = {
    np.dtype('int8'): int8_nan,
    np.dtype('uint8'): uint8_nan,
    np.dtype('int16'): int16_nan,
    np.dtype('uint16'): uint16_nan,
    np.dtype('int32'): int32_nan,
    np.dtype('uint32'): uint32_nan,
    np.dtype('int64'): int64_nan,
    np.dtype('uint64'): uint64_nan,
    np.dtype('float32'): float32_nan,
    np.dtype('float64'): float64_nan,
    np.dtype('<U1'): None,
    np.dtype('O'): None,
}

def mask_dtype_nans(ds):
    """
    Mask NaNs in a dataset based on the dtypes of its variables.
    """
    for var in ds.data_vars:
        if ds[var].dtype in dtype_to_nan and ~ds[var].isnull().any():
            ds[var] = ds[var].where(ds[var] != dtype_to_nan[ds[var].dtype])
    return ds


def get_bit(a, bit_pos):
    """
    Returns 1 or 0 if bit is set or not.

    Parameters
    ----------
    a : int or numpy.ndarray
      Input array.
    bit_pos : int
      Bit position. First bit position is right.

    Returns
    -------
    b : numpy.ndarray
      1 if bit is set and 0 if not.
    """
    return np.clip(np.bitwise_and(a, 2**(bit_pos - 1)), 0, 1)


def set_bit(a, bit_pos, value=1):
    """
    Set bit at given position.

    Parameters
    ----------
    a : int or numpy.ndarray
      Input array.
    bit_pos : int
      Bit position. First bit starts right.
    value : 1 or 0, optional
      Set bit either to 1 or 0 (default: 1).

    Returns
    -------
    a : numpy.ndarray
      Modified input array with bit=value.
    """
    if value == 1:
        return np.bitwise_or(np.atleast_1d(a), 2**(bit_pos - 1))
    else:
        return np.bitwise_and(np.atleast_1d(a), ~(2**(bit_pos - 1)))


def daterange(start_date, end_date):
    """
    Generator for daily datetimes.

    Parameters
    ----------
    start_date : datetime
        Start date.
    end_date : datetime
        End date.
    """
    for n in range(int((end_date - start_date).days)):
        yield start_date + timedelta(n)


def tmp_unzip(filename):
    """
    Unzip file to temporary directory.

    Parameters
    ----------
    filename : str
        Filename.

    Returns
    -------
    unzipped_filename : str
        Unzipped filename
    """
    with NamedTemporaryFile(delete=False) as tmp_fid:
        with GzipFile(filename) as gz_fid:
            tmp_fid.write(gz_fid.read())
        unzipped_filename = tmp_fid.name

    return unzipped_filename


def db2lin(val):
    """
    Converting from linear to dB domain.

    Parameters
    ----------
    val : numpy.ndarray
        Values in dB domain.

    Returns
    -------
    val : numpy.ndarray
        Values in linear domain.
    """
    return 10**(val / 10.)


def lin2db(val):
    """
    Converting from linear to dB domain.

    Parameters
    ----------
    val : numpy.ndarray
        Values in linear domain.

    Returns
    -------
    val : numpy.ndarray
        Values in dB domain.
    """
    return 10. * np.log10(val)


def get_window_radius(window, hp_radius):
    """
    Calculates the required radius of a window function in order to achieve
    the provided half power radius.

    Parameters
    ----------
    window : string
        Window function name.
        Current supported windows:
            - Hamming
            - Boxcar
    hp_radius : float32
        Half power radius. Radius of window function for weight
        equal to 0.5 (-3 dB). In the spatial domain this corresponds to
        half of the spatial resolution one would like to achieve with the
        given window.
    Returns
    -------
    r : float32
        Window radius needed to achieve the given half power radius

    """
    window = window.lower()
    hp_weight = 0.5
    if window == 'hamming':
        alpha = 0.54
        r = (np.pi * hp_radius) / np.arccos((hp_weight - alpha) / (1 - alpha))
    elif window == 'boxcar':
        r = hp_radius
    else:
        raise ValueError('Window name not supported.')

    return r


def hamming_window(radius, distances):
    """
    Hamming window filter.

    Parameters
    ----------
    radius : float32
        Radius of the window.
    distances : numpy.ndarray
        Array with distances.

    Returns
    -------
    weights : numpy.ndarray
        Distance weights.
    tw : float32
        Sum of weigths.
    """
    alpha = 0.54
    weights = alpha + (1 - alpha) * np.cos(np.pi / radius * distances)

    return weights, np.sum(weights)


def boxcar(radius, distance):
    """
    Boxcar filter

    Parameters
    ----------
    n : int
        Length.

    Returns
    -------
    weights : numpy.ndarray
        Distance weights.
    tw : float32
        Sum of weigths.
    """
    weights = np.zeros(distance.size)
    weights[distance <= radius] = 1.

    return weights, np.sum(weights)


def get_window_weights(window, radius, distance, norm=False):
    """
    Function returning weights for the provided window function

    Parameters
    ----------
    window : str
        Window function name
    radius : float
        Radius of the window.
    distance : numpy.ndarray
        Distance array
    norm : boolean
        If true, normalised weights will be returned.

    Returns
    -------
    weights : numpy.ndarray
        Weights according to distances and given window function

    """
    if window == 'hamming':
        weights, w_sum = hamming_window(radius, distance)
    elif window == 'boxcar':
        weights, w_sum = boxcar(radius, distance)
    else:
        raise ValueError('Window name not supported.')

    if norm is True:
        weights = weights / w_sum

    return weights


def get_toi_subset(ds, toi):
    """
    Filter dataset for given time of interest.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to be filtered for time of interest.
    toi : tuple of datetime
        Time of interest.

    Returns
    -------
    ds : xarray.Dataset
        Filtered dataset.
    """
    if isinstance(ds, dict):
        for key in ds.keys():
            subset = np.where((ds[key]['time'] > np.datetime64(toi[0]))
                              & (ds[key]['time'] < np.datetime64(toi[1])))[0]
            if subset.size == 0:
                ds[key] = None
            else:
                if isinstance(ds[key], xr.Dataset):
                    ds[key] = ds[key].sel(obs=np.nonzero(subset.values)[0])
                elif isinstance(ds[key], np.ndarray):
                    ds[key] = ds[key][subset]
    else:
        subset = np.where((ds['time'] > np.datetime64(toi[0]))
                          & (ds['time'] < np.datetime64(toi[1])))[0]
        if subset.size == 0:
            ds = None
        else:
            if isinstance(ds, xr.Dataset):
                ds = ds.sel(obs=np.nonzero(subset.values)[0])
            elif isinstance(ds, np.ndarray):
                ds = ds[subset]

    return ds


def get_roi_subset(ds, roi):
    """
    Filter dataset for given region of interest.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to be filtered for region of interest.
    roi : tuple of 4 float
        Region of interest: latmin, lonmin, latmax, lonmax

    Returns
    -------
    ds : xarray.Dataset
        Filtered dataset.
    """
    if isinstance(ds, dict):
        for key in ds.keys():
            subset = np.where((ds[key]['lat'] > roi[0])
                              & (ds[key]['lat'] < roi[2])
                              & (ds[key]['lon'] > roi[1])
                              & (ds[key]['lon'] < roi[3]))[0]
            if subset.size == 0:
                ds[key] = None
            else:
                if isinstance(ds[key], xr.Dataset):
                    ds[key] = ds[key].sel(obs=np.nonzero(subset.values)[0])
                elif isinstance(ds[key], np.ndarray):
                    ds[key] = ds[key][subset]
    else:
        subset = np.where((ds['lat'] > roi[0]) & (ds['lat'] < roi[2])
                          & (ds['lon'] > roi[1]) & (ds['lon'] < roi[3]))[0]
        if subset.size == 0:
            ds = None
        else:
            if isinstance(ds, xr.Dataset):
                ds = ds.sel(obs=np.nonzero(subset.values)[0])
            elif isinstance(ds, np.ndarray):
                ds = ds[subset]

    return ds

def get_grid_gpis(
        grid,
        cell=None,
        location_id=None,
        coords=None,
        bbox=None,
        geom=None,
        max_coord_dist=np.inf,
        return_lookup: bool = False,
):
    """
    Get grid point indices.

    Parameters
    ----------
    grid : pygeogrids.CellGrid
        Grid object.
    cell : int or iterable of int, optional
        Cell number(s).
    location_id : int or iterable of int, optional
        Location ID.
    coords : tuple, optional
        Tuple of (lon, lat) coordinates.
    bbox : tuple, optional
        Tuple of (latmin, latmax, lonmin, lonmax) coordinates.
    geom : shapely.geometry.BaseGeometry, optional
        Geometry object.
    max_coord_dist : float, optional
        Maximum distance from coordinates to return a gpi.

    Returns
    -------
    gpi : int
        Grid point index.
    lookup_vector : numpy.ndarray
        Lookup vector. (only if return_lookup is True)
    """
    if cell is not None:
        gpis, _, _ = grid.grid_points_for_cell(cell)
    elif location_id is not None:
        gpis = location_id
        if not isinstance(gpis, list) and not isinstance(gpis, np.ndarray):
            gpis = [gpis]
        gpis = np.array(gpis)
    elif coords is not None:
        gpis, dist = grid.find_nearest_gpi(*coords)
        gpis = np.array(gpis)
        dist = np.array(dist)
        gpis = gpis[dist <= max_coord_dist]

    elif bbox is not None:
        gpis = grid.get_bbox_grid_points(*bbox)
    elif geom is not None:
        lonmin, latmin, lonmax, latmax = geom.bounds
        bbox_gpis, bbox_lats, bbox_lons = grid.get_bbox_grid_points(
            latmin,
            latmax,
            lonmin,
            lonmax,
            both=True
        )
        if len(bbox_gpis) > 0:
            gpis = [
                gpi
                for gpi, lat, lon in zip(bbox_gpis, bbox_lats, bbox_lons)
                if geom.contains(Point(lon, lat))
            ]
        else:
            gpis = None
    else:
        gpis = None

    gpis = np.unique(gpis)
    if return_lookup:
        lookup_vector = gpis_to_lookup(grid, gpis)
        return gpis, lookup_vector

    return gpis

def gpis_to_lookup(grid, gpis):
    """
    Create lookup vector from grid point indices.

    Parameters
    ----------
    grid : pygeogrids.BasicGrid
        Grid object.
    gpis : numpy.ndarray
        Grid point indices.

    Returns
    -------
    lookup_vector : numpy.ndarray
        Lookup vector.
    """
    if gpis is not None:
        lookup_vector = np.zeros(grid.gpis.max()+1, dtype=bool)
        lookup_vector[gpis] = 1
        return lookup_vector

def append_to_netcdf(filename, ds_to_append, unlimited_dim):
    """Appends an xarray dataset to an existing netCDF file along a given unlimited dim.

    Parameters
    ----------
    filename : str or Path
        Filename of netCDF file to append to.
    ds_to_append : xarray.Dataset
        Dataset to append.
    unlimited_dim : str or list of str
        Name of the unlimited dimension to append along.

    Raises
    ------
    ValueError
        If more than one unlimited dim is given.
    """
    # By @hmaarrfk on github: https://github.com/pydata/xarray/issues/1672
    with netCDF4.Dataset(filename, mode='a') as nc:
        nc.set_auto_maskandscale(False)
        nc_coord = nc.dimensions[unlimited_dim]
        nc_shape = len(nc_coord)

        added_size = ds_to_append.sizes[unlimited_dim]
        variables, _ = xr.conventions.encode_dataset_coordinates(ds_to_append)

        for name, data in variables.items():
            if unlimited_dim not in data.dims:
                # Nothing to do, data assumed to the identical
                continue

            nc_variable = nc[name]
            _expand_variable(nc_variable, data, unlimited_dim, nc_shape,
                             added_size)

def _expand_variable(nc_variable, data, expanding_dim, nc_shape, added_size):
    # Adapted from @hmaarrfk on github: https://github.com/pydata/xarray/issues/1672
    # For time deltas, we must ensure that we use the same encoding as
    # what was previously stored.
    # We likely need to do this as well for variables that had custom
    # encodings too
    data.encoding = dict()
    if hasattr(nc_variable, 'calendar'):
        data.encoding['calendar'] = nc_variable.calendar
    if hasattr(nc_variable, 'calender'):
        data.encoding['calendar'] = nc_variable.calender

    if hasattr(nc_variable, 'dtype'):
        data.encoding['dtype'] = nc_variable.dtype
    if hasattr(nc_variable, 'units'):
        data.encoding['units'] = nc_variable.units
    if hasattr(nc_variable,
               '_FillValue') and data.attrs.get('_FillValue') is None:
        data.encoding['_FillValue'] = nc_variable._FillValue
    if hasattr(nc_variable,
               'missing_value') and data.attrs.get('missing_value') is None:
        data.encoding['missing_value'] = nc_variable.missing_value
    if hasattr(nc_variable,
               'scale_factor') and data.attrs.get('scale_factor') is None:
        data.encoding['scale_factor'] = nc_variable.scale_factor

    data_encoded = xr.conventions.encode_cf_variable(data)

    left_slices = data.dims.index(expanding_dim)
    right_slices = data.ndim - left_slices - 1
    nc_slice = ((slice(None),) * left_slices +
                (slice(nc_shape, nc_shape + added_size),) + (slice(None),) *
                (right_slices))
    nc_variable[nc_slice] = data_encoded.data

def create_variable_encodings(ds,
                              custom_variable_encodings=None,
                              custom_dtypes=None):
    """
    Create an encoding dictionary for a dataset, optionally
    overriding the default encoding or adding additional
    encoding parameters.
    New parameters cannot be added to default encoding for
    a variable, only overridden.

    E.g. if you want to add a "units" encoding to "lon",
    you should also pass "dtype", "zlib", "complevel",
    and "_FillValue" if you don't want to lose those.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset.
    custom_variable_encodings : dict, optional
        Custom encodings.

    Returns
    -------
    ds : xarray.Dataset
        Dataset with encodings.
    """
    if custom_variable_encodings is None:
        custom_variable_encodings = {}

    if "row_size" in ds.data_vars:
        # contiguous RA case
        first_var = {"row_size": {"dtype": "int64"}}
        coord_vars = {
            "lon": {"dtype": "float32",},
            "lat": {"dtype": "float32",},
            "alt": {"dtype": "float32",},
        }
    elif "locationIndex" in ds.data_vars:
        # indexed RA case
        first_var = {"locationIndex": {"dtype": "int64"}}
        coord_vars = {
            "lon": {"dtype": "float32"},
            "lat": {"dtype": "float32"},
            "alt": {"dtype": "float32"},
        }
    else:
        # swath file case
        first_var = {}
        coord_vars = {
            "lon": {"dtype": "float32"},
            "lat": {"dtype": "float32"},
            # "longitude": {"dtype": "float32"},
            # "latitude": {"dtype": "float32"},
        }

    # default encodings for coordinates and row_size
    default_encoding = {
        **first_var,
        **coord_vars,
        "location_id": {
            "dtype": "int64",
        },
        "time": {
            "dtype": "float64",
            "units": "days since 1900-01-01 00:00:00",
        },
    }

    for _, var_encoding in default_encoding.items():
        if var_encoding["dtype"] != "int64":
            var_encoding["_FillValue"] = None
        var_encoding["zlib"] = True
        var_encoding["complevel"] = 4

    for var, dtype in ds.dtypes.items():
        if var not in default_encoding:
            dtype = ds[var].encoding.get("dtype", dtype)
            default_encoding[var] = {
                "dtype": dtype,
                "zlib": bool(np.issubdtype(dtype, np.number)),
                "complevel": 4,
                "_FillValue": dtype_to_nan[dtype],
            }

    if custom_dtypes is not None:
        custom_variable_encodings = {
            var: {
                "dtype": custom_dtypes[var],
                "zlib": bool(np.issubdtype(custom_dtypes[var], np.number)),
                "complevel": 4,
                "_FillValue": dtype_to_nan[custom_dtypes[var]],
            } for var in custom_dtypes.names if var in ds.data_vars
        }

    encoding = {**default_encoding, **custom_variable_encodings}

    return encoding
def get_file_format(filename):
    """
    Try to guess the file format from the extension.

    Parameters
    ----------
    filename : str
        File name.

    Returns
    -------
    file_format : str
        File format indicator.
    """
    if os.path.splitext(filename)[1] == ".gz":
        file_format = os.path.splitext(os.path.splitext(filename)[0])[1]
    else:
        file_format = os.path.splitext(filename)[1]

    return file_format

class Spacecraft:
    """
    Spacecraft class.
    """

    valid_spacecraft_names = [
        "METOPA", "METOPB", "METOPC", "METOP-A", "METOP-B", "METOP-C",
        "METOP-SG B1", "METOP-SG B2", "METOP-SG B3"
    ]

    def __init__(self, name):
        """
        Initialize spacecraft class.

        Parameters
        ----------
        name : str
            Spacecraft name.
        """
        if name not in Spacecraft.valid_spacecraft_names:
            valid_names = ' ,'.join(Spacecraft.valid_spacecraft_names)
            msg = f"Spacecraft {name} unknown. Valid options: {valid_names}"
            raise RuntimeError(msg)

        satellite_dict = {
            "METOPA": "METOPA",
            "METOPB": "METOPB",
            "METOPC": "METOPC",
            "METOP-A": "METOPA",
            "METOP-B": "METOPB",
            "METOP-C": "METOPC",
            "METOP-SG B1": "METOP-SGB1",
            "METOP-SG B2": "METOP-SGB2",
            "METOP-SG B3": "METOP-SGB3"
        }

        platform_dict = {
            "METOPA": "Metop",
            "METOPB": "Metop",
            "METOPC": "Metop",
            "METOP-SGB1": "Metop-SG",
            "METOP-SGB2": "Metop-SG",
            "METOP-SGB3": "Metop-SG"
        }

        sensor_dict = {"Metop": "ASCAT", "Metop-SG": "SCA"}

        sat_name_dict = {
            "METOPA": "A",
            "METOPB": "B",
            "METOPC": "C",
            "METOP-SGB1": "B1",
            "METOP-SGB2": "B2",
            "METOP-SGB3": "B3"
        }

        sat_id_dict = {
            "METOPA": 3,
            "METOPB": 4,
            "METOPC": 5,
            "METOP-SGB1": 6,
            "METOP-SGB2": 7,
            "METOP-SGB3": 8
        }

        self.satellite = satellite_dict[name]
        self.platform = platform_dict[self.satellite]
        self.sensor = sensor_dict[self.platform]
        self.sat_name = sat_name_dict[self.satellite]
        self.sat_id = sat_id_dict[self.satellite]



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import argparse
from pathlib import Path

import ascat.aggregate.aggregators as aggs
from ascat.regrid.interface import swath_regrid_main
from ascat.resample.interface import swath_resample_main


def parse_args_temporal_swath_agg(args):
    parser = argparse.ArgumentParser(
        description=("Generate aggregates of ASCAT swath data over "
                     "a given time period"))
    parser.add_argument(
        "filepath", metavar="FILEPATH", help="Path to the data")
    parser.add_argument(
        "outpath", metavar="OUTPATH", help="Path to the output data")
    parser.add_argument(
        "--start_dt",
        metavar="START_DT",
        help="Start datetime (formatted e.g. 2020-01-01T00:00:00)")
    parser.add_argument(
        "--end_dt",
        metavar="END_DT",
        help="End datetime (formatted e.g. 2020-02-01T00:00:00)")
    parser.add_argument(
        "--t_delta",
        metavar="T_DELTA",
        help="Time period for aggregation (e.g. 1D, 1W, 1M, 1Y, 2D, etc.)")
    parser.add_argument("--agg", metavar="AGG", help="Aggregation")
    parser.add_argument(
        "--sat",
        metavar="SAT",
        type=str,
        default="*",
        help=("Regex pattern indicating which METOP satellite(s) to use, "
              " e.g. A, [BC], or '*' "
              "(default: * for all satellites)"))
    parser.add_argument(
        "--snow_cover_mask",
        metavar="SNOW_COVER_MASK",
        type=int,
        default=90,
        help=("Snow cover probability (0-100 %) value above which "
              "to mask the source data (default: 90 %)"))
    parser.add_argument(
        "--frozen_soil_mask",
        metavar="FROZEN_SOIL_MASK",
        type=int,
        default=90,
        help=("Frozen soil probability (0-100 %) value above which "
              "to mask the source data (default: 90 %)"))
    parser.add_argument(
        "--subsurface_scattering_mask",
        metavar="SUBSURFACE_SCATTERING_MASK",
        type=int,
        default=10,
        help=("Subsurface scattering probability (0-100 %) value above which "
              "to mask the source data (default: 10 %)"))
    parser.add_argument(
        "--ssm_sensitivity_mask",
        metavar="SSM_SENSITIVITY_MASK",
        type=float,
        default=1.0,
        help=("Surface soil moisture sensitivity (in dB) value below which "
              "to mask the source data (default: 1 dB)"))
    parser.add_argument(
        "--no_masking",
        action='store_const',
        const=True,
        default=False,
        help="Ignore all masks")
    parser.add_argument(
        "--regrid",
        metavar="REGRID_DEG",
        type=float,
        help=("Regrid the data to a regular grid with the given "
              " spacing in degrees"))
    parser.add_argument(
        "--resample",
        metavar="RESAMPLE_DEG",
        type=float,
        help=("Resample the data to a regular grid with the given "
              " spacing in degrees"))
    parser.add_argument(
        "--resample_neighbours",
        metavar="RESAMPLE_NEIGHBOURS",
        type=int,
        default=6,
        help="Number of neighbours to consider for each grid point when resampling (default: 6)")
    parser.add_argument(
        "--resample_radius",
        metavar="RESAMPLE_RADIUS",
        type=float,
        default=10000,
        help="Cut off distance in meters (default: 10000)")
    parser.add_argument(
        "--grid_store",
        metavar="GRID_STORE",
        help=("Path to a directory for storing grids and "
              "lookup tables between them"))
    parser.add_argument(
        "--suffix",
        metavar="SUFFIX",
        help="File suffix (default: _REGRID_DEGdeg)")

    return parser.parse_args(args)


def temporal_swath_agg_main(cli_args):
    """
    Command line interface routine for temporal aggregation of ASCAT data.

    Parameters
    ----------
    cli_args : list
        Command line arguments.
    """
    args = parse_args_temporal_swath_agg(cli_args)

    transf = aggs.TemporalSwathAggregator(
        args.filepath, args.start_dt, args.end_dt, args.t_delta, args.agg,
        args.snow_cover_mask, args.frozen_soil_mask,
        args.subsurface_scattering_mask, args.ssm_sensitivity_mask,
        args.no_masking, sat=args.sat)

    product_id = transf.product

    outpath = Path(args.outpath)
    outpath.mkdir(parents=True, exist_ok=True)

    filenames = transf.write_time_steps(outpath)

    if args.regrid is not None:
        for filename in filenames:
            regrid_args = [str(filename), str(outpath), str(args.regrid)]
            if args.grid_store is not None:
                regrid_args.extend(["--grid_store", args.grid_store])
            if args.suffix is not None:
                regrid_args.extend(["--suffix", args.suffix])
            regrid_args.extend(["--product_id", product_id])
            swath_regrid_main(regrid_args)

    if args.resample is not None:
        for filename in filenames:
            resample_args = [str(filename), str(outpath), str(args.resample)]
            if args.grid_store is not None:
                resample_args.extend(["--grid_store", args.grid_store])
            if args.suffix is not None:
                resample_args.extend(["--suffix", args.suffix])
            resample_args.extend(["--neighbours", str(args.resample_neighbours)])
            resample_args.extend(["--radius", str(args.resample_radius)])
            resample_args.extend(["--product_id", product_id])
            swath_resample_main(resample_args)

def run_temporal_swath_agg():
    """
    Run command line interface for temporal aggregation of ASCAT data.
    """
    temporal_swath_agg_main(sys.argv[1:])



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from pathlib import Path


def get_current_path():
    """
    Get current file path.
    """
    return Path(__file__).resolve().parent


def get_testdata_path():
    """
    Get test data path.
    """
    return Path(__file__).resolve().parent / "ascat-test-data"



#!/usr/bin/env python3

import unittest
from pathlib import Path
from tempfile import TemporaryDirectory

import xarray as xr
import numpy as np


from ascat.read_native import generate_test_data as gtd

# from ascat.read_native.product_info import cell_io_catalog

from ascat.accessors import *
from ascat.cell import RaggedArrayTs


def add_sat_id(ds, sat_name):
    name_dict = {"metop_a": 3, "metop_b": 4, "metop_c": 5}
    if sat_name is not None:
        ds["sat_id"] = ("time", np.repeat(name_dict[sat_name], ds.sizes["time"]))
    return ds


def gen_dummy_cellfiles(dir, sat_name=None):
    contiguous_dir = dir / "contiguous"
    if sat_name is not None:
        contiguous_dir = contiguous_dir / sat_name
    contiguous_dir.mkdir(parents=True, exist_ok=True)
    indexed_dir = dir / "indexed"
    if sat_name is not None:
        indexed_dir = indexed_dir / sat_name
    indexed_dir.mkdir(parents=True, exist_ok=True)
    add_sat_id(gtd.contiguous_ragged_ds_2588, sat_name).to_netcdf(
        contiguous_dir / "2588.nc"
    )
    add_sat_id(gtd.indexed_ragged_ds_2588, sat_name).to_netcdf(indexed_dir / "2588.nc")
    add_sat_id(gtd.contiguous_ragged_ds_2587, sat_name).to_netcdf(
        contiguous_dir / "2587.nc"
    )
    add_sat_id(gtd.indexed_ragged_ds_2587, sat_name).to_netcdf(indexed_dir / "2587.nc")


class TestCFDiscreteGeometryAccessor(unittest.TestCase):
    @staticmethod
    def _init_options(root_path, sf_templ=None, sf_read_fmt=None):
        return {
            "root_path": root_path,
            "file_class": RaggedArrayTs,
            "fn_format": "{:04d}.nc",
            # "sf_format": sf_templ,
        }

    def setUp(self):
        self.tempdir = TemporaryDirectory()
        self.tempdir_path = Path(self.tempdir.name)
        gen_dummy_cellfiles(self.tempdir_path)
        gen_dummy_cellfiles(self.tempdir_path, "metop_a")
        gen_dummy_cellfiles(self.tempdir_path, "metop_b")
        gen_dummy_cellfiles(self.tempdir_path, "metop_c")

    def tearDown(self):
        self.tempdir.cleanup()

    def test_array_type(self):
        ds = xr.open_dataset(self.tempdir_path / "contiguous" / "2588.nc")

        ds.attrs["grid_mapping_name"] = "fibgrid_12.5"
        ds["row_size"] = ds.row_size.assign_attrs({"sample_dimension": "time"})
        ds["location_id"] = ds["location_id"].assign_attrs({"cf_role": "timeseries_id"})

        # ds = ds.set_coords(["lat", "lon", "alt"])
        # ds.cf_geom.set_coord_vars(["lat", "lon", "alt"])
        # assert ds.cf_geom._coord_vars == ["lat", "lon", "alt"]

        assert ds.cf_geom.array_type == "contiguous"

        # to indexed_ragged
        ids = ds.cf_geom.to_indexed_ragged()
        assert ids.cf_geom.array_type == "indexed"

        # to indexed_ragged with different index_var name
        ids2 = ds.copy().cf_geom.to_indexed_ragged(index_var="BIGTEST")
        assert ids2.cf_geom.array_type == "indexed"

        # now back to_contiguous_ragged
        cds = ids2.cf_geom.to_contiguous_ragged()
        assert cds.cf_geom.array_type == "contiguous"
        xr.testing.assert_identical(cds, ds)

        # now back to indexed
        ids3 = cds.cf_geom.to_indexed_ragged()
        assert ids3.cf_geom.array_type == "indexed"
        assert ids3.cf_geom._obj._index_var != "BIGTEST"

        # original to point array
        pads = ds.cf_geom.to_point_array()
        assert pads.cf_geom.array_type == "point"


class TestPyGeoGriddedArrayAccessor(unittest.TestCase):
    @staticmethod
    def _init_options(root_path, sf_templ=None, sf_read_fmt=None):
        return {
            "root_path": root_path,
            "file_class": RaggedArrayTs,
            "fn_format": "{:04d}.nc",
            # "sf_format": sf_templ,
        }

    def setUp(self):
        self.tempdir = TemporaryDirectory()
        self.tempdir_path = Path(self.tempdir.name)
        gen_dummy_cellfiles(self.tempdir_path)
        gen_dummy_cellfiles(self.tempdir_path, "metop_a")
        gen_dummy_cellfiles(self.tempdir_path, "metop_b")
        gen_dummy_cellfiles(self.tempdir_path, "metop_c")

    def tearDown(self):
        self.tempdir.cleanup()

    def test_array_type(self):
        ds = xr.open_dataset(self.tempdir_path / "contiguous" / "2588.nc")

        ds.attrs["grid_mapping_name"] = "fibgrid_12.5"
        ds["row_size"] = ds.row_size.assign_attrs({"sample_dimension": "time"})
        ds["location_id"] = ds["location_id"].assign_attrs({"cf_role": "timeseries_id"})


        # select from original contiguous
        gpi_selection = ds.pgg.sel_gpis([1549346, 1556056])
        assert gpi_selection.cf_geom.array_type == ds.cf_geom.array_type
        np.testing.assert_array_equal(
            gpi_selection["location_id"].values, [1549346, 1556056]
        )

        # select from indexed
        idx_gpi_selection = (
            ds.load().cf_geom.to_indexed_ragged().pgg.sel_gpis([1549346, 1556056])
        )
        assert idx_gpi_selection.cf_geom.array_type == "indexed"
        xr.testing.assert_identical(
            idx_gpi_selection.cf_geom.to_point_array(),
            gpi_selection.cf_geom.to_point_array(),
        )

        # select from point array
        point_gpi_selection = ds.cf_geom.to_point_array().pgg.sel_gpis(
            [1549346, 1556056]
        )
        assert point_gpi_selection.cf_geom.array_type == "point"

        # contig_ra = RaggedArray(ds)
        # self.assertEqual(contig_ra.array_type, "contiguous")



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Tests for reading CGLOPS SWI data.
"""

import pandas as pd
import numpy as np
import pytest

from get_path import get_testdata_path
from ascat.cgls import SWI_TS


def test_swi_ts_reader():
    """
    Test SWI time series reader.
    """
    data_path = get_testdata_path() / "cglops" / "swi_ts"

    rd = SWI_TS(data_path)
    data = rd.read(3002621, mask_frozen=False)
    data_sorted = data.sort_index()

    assert np.all(data_sorted.index == data.index)
    # just check if enough data is there
    reference_index = pd.date_range('20070101T12:00:00', '20161231T12:00:00')
    assert len(data) == len(reference_index)
    assert np.all(data_sorted.index == reference_index)

    lon, lat = rd.grid.gpi2lonlat(3002621)
    data = rd.read(lon, lat, mask_frozen=False)
    data_sorted = data.sort_index()

    assert np.all(data_sorted.index == data.index)
    # just check if enough data is there
    reference_index = pd.date_range('20070101T12:00:00', '20161231T12:00:00')
    assert len(data) == len(reference_index)
    assert np.all(data_sorted.index == reference_index)


def test_swi_ts_reader_no_data_in_folder():
    """
    Test SWI time series reader when no data is in folder.
    """
    data_path = get_testdata_path() / "cglops" / "swi_ts_non_existing"

    with pytest.raises(IOError):
        SWI_TS(data_path)


def test_swi_ts_qflag_reading():
    """
    Test SWI time series quality flag reader.
    """
    data_path = get_testdata_path() / "cglops" / "swi_ts"
    rd = SWI_TS(data_path, parameters=['SWI_001', 'QFLAG_001', 'SSF'])
    data = rd.read(3002621, mask_frozen=True)
    # check if QFLAG is correctly read. It should have as many NaN values as
    # SWI
    assert len(data[data.loc[:, 'QFLAG_001'] != np.nan]) > 0
    assert (len(data[data.loc[:, 'QFLAG_001'] == np.nan]) == len(
        data[data.loc[:, 'SWI_001'] == np.nan]))


if __name__ == "__main__":
    test_swi_ts_reader()
    test_swi_ts_qflag_reading()
    test_swi_ts_reader_no_data_in_folder()



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
CGLS SWI interface.
"""

import os
import glob
import numpy as np

import pynetcf.time_series as netcdf_dataset
import pygeogrids.netcdf as netcdf


class SWI_TS(netcdf_dataset.GriddedNcOrthoMultiTs):

    """
    SWI TS reader for timeseries data from CGLOPS

    Parameters
    ----------
    data_path : string
        path to the netCDF files
    parameters : list
        list of parameters to read from netCDF file
    dt : string, optional
        datetime in the filenames of the cells.
        If not given it is detected from the files in the data_path.
        Automatic detection only works if the files follow the CGLS
        naming convention.
    version : string, optional
        version number of the files
        If not given it is detected from the files in the data_path.
        Automatic detection only works if the files follow the CGLS
        naming convention.
    grid_fname : string, optional
        filename + path of the grid netCDF file,
        default is the standard grid file
        (c_gls_SWI-STATIC-DGG_201501010000_GLOBE_ASCAT_V3.0.1.nc)
        in the same folder as the data
    read_bulk : boolean, optional
        if set to true then a complete 5x5 degree cell will be read at once
        providing speedup if the complete data is needed.
    fname_template : string, optional
        Filename template. Has to have three slots for {dt}, {version} and a
        slot for the {cell} number that is available for further formatting.
        The has to be without the .nc ending since this is added during reading.
    cell_fn : string, optional
        cell number in the fname_template.
    """

    def __init__(self, data_path, parameters=['SWI_001', 'SWI_005', 'SWI_010',
                                              'SWI_015', 'SWI_020', 'SWI_040',
                                              'SWI_060', 'SWI_100', 'SSF'],
                 dt=None, version=None,
                 grid_fname=None, read_bulk=True,
                 fname_template='c_gls_SWI-TS_{dt}_C{cell}_ASCAT_V{version}',
                 cell_fn='{:04d}'):

        if grid_fname is None:
            grid_fname = os.path.join(
                data_path, 'c_gls_SWI-STATIC-DGG_201501010000_GLOBE_ASCAT_V3.0.1.nc')
        grid = netcdf.load_grid(grid_fname, location_var_name='location_id',
                                subset_flag='land_flag')

        # detect datetime and version if not given
        if dt is None or version is None:
            globstring = fname_template.format(dt="*",
                                               cell="*",
                                               version="*")
            found_files = glob.glob(os.path.join(data_path, globstring))
            if len(found_files) == 0:
                raise IOError("No data found in {}".format(data_path))
            fn = found_files[0]
            fn = os.path.splitext(os.path.basename(fn))[0]
            parts = fn.split('_')
        if dt is None:
            # this only works if the files follow the CGLS naming convention
            # for everything else dt should be given as a keyword
            dt = parts[3]
        if version is None:
            version = parts[-1][1:]

        scale_factors = {'SWI_001': 0.5,
                         'SWI_005': 0.5,
                         'SWI_010': 0.5,
                         'SWI_015': 0.5,
                         'SWI_020': 0.5,
                         'SWI_040': 0.5,
                         'SWI_060': 0.5,
                         'SWI_100': 0.5,
                         'QFLAG_001': 0.5,
                         'QFLAG_005': 0.5,
                         'QFLAG_010': 0.5,
                         'QFLAG_015': 0.5,
                         'QFLAG_020': 0.5,
                         'QFLAG_040': 0.5,
                         'QFLAG_060': 0.5,
                         'QFLAG_100': 0.5,
                         'SSF': 1}

        dtypes = {'SWI_001': np.uint8,
                  'SWI_005': np.uint8,
                  'SWI_010': np.uint8,
                  'SWI_015': np.uint8,
                  'SWI_020': np.uint8,
                  'SWI_040': np.uint8,
                  'SWI_060': np.uint8,
                  'SWI_100': np.uint8,
                  'QFLAG_001': np.uint8,
                  'QFLAG_005': np.uint8,
                  'QFLAG_010': np.uint8,
                  'QFLAG_015': np.uint8,
                  'QFLAG_020': np.uint8,
                  'QFLAG_040': np.uint8,
                  'QFLAG_060': np.uint8,
                  'QFLAG_100': np.uint8,
                  'SSF': np.uint8}

        super(SWI_TS, self).__init__(
            data_path, grid,
            fn_format=fname_template.format(dt=dt, version=version,
                                            cell=cell_fn),
            parameters=parameters, scale_factors=scale_factors,
            dtypes=dtypes, autoscale=False,
            automask=False, ioclass_kws={'read_bulk': read_bulk,
                                         'loc_ids_name': 'locations'})

    def _read_gp(self, gpi, period=None, mask_frozen=True):
        data = super(SWI_TS, self)._read_gp(gpi, period=period)

        if mask_frozen is True:
            unfrozen = data['SSF'].values <= 1
            data = data[unfrozen]

        for column in data:
            data.loc[data[column] > 100, column] = np.nan

        return data



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from enum import Enum
from pathlib import Path
from typing import Dict, Type, Any, Tuple

from fibgrid.realization import FibGrid
from pygeogrids.netcdf import load_grid


class GridType(Enum):
    FIBGRID = "fibgrid"
    NAMED = "named"


class SingletonArgs(type):
    _instances: Dict[Tuple[Type, Tuple, frozenset], Any] = {}

    def __call__(cls, *args, **kwargs):
        key = (cls, args, frozenset(kwargs.items()))
        if key not in cls._instances:
            cls._instances[key] = super().__call__(*args, **kwargs)
        return cls._instances[key]


class GridSingleton(metaclass=SingletonArgs):
    __slots__ = ['grid']

    def __init__(self, grid_class: Type, *args):
        self.grid = grid_class(*args)


class NamedFileGridRegistry:
    _grids: Dict[str, str] = {}

    @classmethod
    def register(cls, grid_name: str, grid_path: str) -> None:
        """Register a named grid with its file path."""
        cls._grids[grid_name] = Path(grid_path)

    @classmethod
    def get(cls, grid_name: str) -> str:
        """Retrieve the file path for a registered grid."""
        if grid_name not in cls._grids:
            raise KeyError(f"Grid '{grid_name}' is not registered.")
        return cls._grids[grid_name]


class NamedFileGrid:

    def __new__(cls, grid_name: str):
        grid_path = NamedFileGridRegistry.get(grid_name)
        return load_grid(grid_path)


class GridRegistry:
    _registry = {
        "fibgrid": FibGrid,
        "named": NamedFileGrid,
    }

    def register(
        self,
        grid_type_name: str,
        grid_class: type,
    ):
        """
        Register a grid class with a name for later retrieval.

        e.g. `register("fibgrid", FibGrid)` or `register("named", NamedFileGrid)`
        """
        if grid_type_name in self._registry:
            return
        self._registry[grid_type_name] = grid_class

    def get(self, grid_name):
        """
        Retrieve a grid instance based on its name.

        The grid name can be a simple name (e.g. "fibgrid") or a more complex
        name with parameters (e.g. "fibgrid_0.1"). The latter will be split
        into the grid type and its parameters.

        Parameters
        ----------
            grid_name (str): The name of the grid to retrieve.
        """
        parts = grid_name.split("_")

        if len(parts) >= 2 and parts[0] == "fibgrid":
            grid_type = "fibgrid"
            grid_spacing = float(parts[1])
            args = (grid_spacing,)
        elif len(parts) == 1:
            grid_type = "named"
            args = (grid_name,)
        else:
            grid_type = parts[0]
            args = tuple(parts[1:])

        grid_class = self._registry.get(grid_type)
        if grid_class is None:
            raise KeyError(f"Grid {grid_name} is not registered.")

        return GridSingleton(grid_class, *args).grid
